<?php

//Constants for use with the family of NaiveSqlParser::eat* functions. 
define('WS_SQL_RETURN_COUNT', 0);
define('WS_SQL_RETURN_ARRAY', 1);
define('WS_SQL_RETURN_CONCATENATED', 2);

/**
 * Primitive SQL parser capable of analyzing simple CREATE TABLE queries.
 * 
 * @author Janis Elsts
 * @copyright 2010
 * @version 1.0
 * @access public
 */
class NaiveSqlParser {
	var $sql = ''; //The input string containing one or more SQL queries.  
	
	var $tokens;   //Tokenized version of the input string.
	var $queries;  //Array of queries parsed so far.
	
	var $num_tokens = 0; //Total number of tokens found in the input string.
	var $index = -1;     //The position of the next token to be analyzed.
	var $_next_salient_index = -1; //The position of the next non-whitespace, non-comment token.
	var $_not_whitespace;          //Array of flags used to quickly identify (non-)whitespace tokens.  
	
	/**
	 * Class constructor.
	 * 
	 * @param string $sql The query/-ies to parse. Alternatively, the input string can also be specified as an argument to the parse() method.
	 * @return void
	 */
	function __construct($sql = ''){
		$this->sql = $sql;
	}
	
	/**
	 * PHP4-style constructor.
	 * 
	 * @see NaiveSqlParser::__construct()
	 * 
	 * @param string $sql
	 * @return void
	 */
	function NaiveSqlParser($sql = ''){
		$this->__construct($sql);
	}
	
	/**
	 * Parse one or more SQL queries.
	 * 
	 * Only CREATE TABLE queries are supported. All other queries will be returned unparsed.
	 * This function returns an array of parsed query structures. Each item is an associated 
	 * array with these keys : 
	 * 	'query_type' - type of the parsed query. Either 'create_table' or 'unknown'.
	 *  'raw_query'  - the substring of the input string corresponding to this query.
	 *  'name'       - the table table name, unquoted.
	 *  'columns'    - array of table columns specified in the CREATE TABLE query, and their attributes.
	 *  'indices'    - array of indices specified in the CREATE TABLE query, and their attributes. 
	 *  'options'    - array of table options.
	 * 
	 * @param string $sql A string containing one or more queries. Each query is expected to be terminated by a semicolon. 
	 * @return array
	 */
	function parse($sql = null){
		if ( isset($sql) && is_string($sql) ){
			$this->sql = $sql;
		}
		if ( empty($this->sql) ){
			return array();
		}
		$this->sql .= ';'; //We expect all queries to be terminated by a semicolon, even if there's only one query.
		
		$this->tokenize();
		
		$this->queries = array();
		while( ($token = $this->nextToken()) !== null ){
			$token_lowercase = strtolower($token);
			$first_token_index = $this->index - 1;
			$parsed_query = null;
			
			switch($token_lowercase) {
				case 'create':
					$what = strtolower($this->lookAhead());
					//Only "CREATE TABLE" queries are supported
					if ( $what == 'table' ){
						$parsed_query = $this->parseCreateTableQuery();
						$parsed_query['query_type'] = 'create_table';
												
						break;
					} else {
						//fall through
					}
					
				default:
					//If the input string already included a trailing semicolon the extra ";"
					//we added is redundant and can be thrown away. This will also discard 
					//"real" no-op queries that consist only of a semicolon. 
					if ( trim($token) == ';' ){
						break;
					}
									
					//Other unsupported queries are turned back into strings and left alone. 
					$remainder = $this->eatUntilEquals(';', WS_SQL_RETURN_ARRAY); //BUG: May stop at e.g. "DEFAULT 'foo;bar'"
					$parsed_query = array(
						'query_type' => 'unknown',
						'raw_query' => $token . implode('', $remainder) . ';',
					);
					break;
			}
			
			if ( $parsed_query ){
				//Having access to the original query can be handy if you need to execute or display it (e.g. for debugging).
				//Note that the array_slice() call below relies on the current index pointing to the next token
				//*after* the end of the current query.
				if ( !isset($parsed_query['raw_query']) ){
					$parsed_query['raw_query'] = implode('', array_slice($this->tokens, $first_token_index, $this->index - $first_token_index));
				}
				$this->queries[] = $parsed_query;
			}
		}	
		
		return $this->queries;
	}
	
	/**
	 * Parse a "CREATE TABLE" query.
	 * 
	 * @return array
	 */
	function parseCreateTableQuery(){
		//Right now the internal pointer should be on "TABLE", so we skip to the next token
		$this->nextToken();
		//Skip the "IF NOT EXISTS" flag, if present
		$this->eatSequence('IF NOT EXISTS');
		//Extract table `name`
		$table = array('name' => $this->parseName());
		
		$this->eatUntilEquals('(');
		
		//Parse all create definitions
		$table['columns'] = array();
		$table['indices'] = array();
		while( $type_and_def = $this->parseCreateDefinition() ){
			list($type, $definition) = $type_and_def;
			if ( $type == 'column' ){
				$table['columns'][$definition['name']] = $definition;
			} elseif ( $type == 'index' ){
				$table['indices'][] = $definition; //TODO: Index indices by the stringified index definition
			} else {
				$table[$type][] = $definition;
			}
		}
		$this->eatUntilEquals(')');
		
		//Parse table options, if any (consumes all tokens remaining in the current query)
		$table['options'] = $this->parseTableOptions();
		
		return $table;
	}
	
	/**
	 * Parse a create definition. 
	 * 
	 * Only column or index definitions are supported. Anything else - i.e. CONSTRAINT's 
	 * and CHECKS's - will lead to garbled output. 
	 * 
	 * @return array Definition type ('column' or 'index') and the definition itself.
	 */
	function parseCreateDefinition(){
		$token = $this->lookAhead();
		if ( $token == ')' || is_null($token) ){
			return null;
		}
		
		$token = strtolower($token);
		$index_keywords = array('primary', 'fulltext', 'unique', 'index', 'key');

		if ( in_array($token, $index_keywords) ) {
			return array('index', $this->parseIndexDefinition());
		} else {
			return array('column', $this->parseColumnDefinition());
		}	
	}
	
	/**
	 * Parse a column definition.
	 * 
	 * @return array
	 */
	function parseColumnDefinition(){
		$name = $this->parseName();
		$data_type = $this->parseDataType();
		
		//Parse flags that apply to all data types - NOT NULL and friends.
		$done = false;
		do {
			$token = strtolower($this->lookAhead());
			switch ( $token ){
				case 'not': //The only flag that begins with "NOT" is NOT NULL
					$this->nextToken(); //eat NOT
					$this->nextToken(); //eat NULL
					$not_null = true;
					break;
					
				case 'null': //The NULL flag is usually not specified explicitly, but it can be.  
					$this->nextToken();
					$not_null = false;
					break;
					
				case 'auto_increment':
					$this->nextToken();
					$auto_increment = true;
					break;
					
				case 'default';
					$this->nextToken();
					$default = strtolower($this->nextToken());
					if ( ($default == '\'') || ($default == '"') ){
						//Parse a quoted string
						$default = $this->parseQuotedString($default);
					} elseif ( $default == 'null' ){
						//DEFAULT NULL is valid, special case
						$default = null;
					}
					//BUG: The above does not differentiate between "DEFAULT 'CURRENT_TIMESTAMP'" (normal string) 
					//     and "DEFAULT CURRENT_TIMESTAMP" (special keyword).
					
					break;
					
				case 'comment':
					$this->nextToken();
					$comment = $this->parseQuotedString(); //Note: Comments *must* be quoted
					break;

				case ',': //Eat the comma that separates create definitions
					$this->nextToken(); 
				case ')': //But the closing paren on "CREATE TABLE xyz ()" is handled in self::parse()
					$done = true;
					break;
					
				default: //Unknown tokens are simply skipped 
					$this->nextToken();
			}			
		} while ( !$done );
		
		$definition = compact('name', 'default', 'not_null', 'auto_increment', 'comment');
		$definition = array_merge($definition, $data_type);
		
		return $definition;
	}
	
	/**
	 * Parse a column's data type specification.
	 * 
	 * Returns an array with these keys: 
	 *   'data_type' - normalized data type, e.g. 'int(10) unsigned'. This matches the format of the 'Type' column returned by the 'DESCRIBE table_name' query.  
	 *   'data_type_name' - the basic datatype name, without length or flags. Example: 'int'.
	 *   'length' - column length, if specified.
	 *   'decimals' - as above.
	 *   'enum_vaues' - an array of possible values for an ENUM column. 
	 *   'unsigned' - True if the 'UNSIGNED' flag was found.
	 *   'zerofill' - likewise.
	 *   'binary'   - likewise.
	 *   'character_set' - the column's character set, if specified.
	 *   'collation' - the column's collation, if specified.
	 * 
	 * Only the 'data_type' and 'data_type_name' keys are guaranteed to be set. All others will 
	 * only be included if the corresponding keywords are present in the column definition.
	 * 
	 * @return array
	 */
	function parseDataType(){
		$data_type_name = $data_type = strtolower($this->nextToken());
		
		//Read optional length or enum value list
		if ( $this->lookAhead() == '(' ){
			$this->nextToken(); //eat the opening paren
			
			if ( ($data_type_name == 'enum') || ($data_type_name == 'set') ){
				//Read the value list. Per MySQL docs, all values must be quoted.
				$enum_values = array();
				do {
					$enum_values[] = $this->parseQuotedString();
				} while ( $this->nextToken() != ')' );
				$data_type .= "('" . implode($enum_values, "','") . "')";
			} else {
				//Read length 
				$length = intval($this->nextToken());
				$data_type .= '(' . $length;
				//And possibly decimals
				if ( $this->nextToken() == ',' ){ //eats the closing paren if there's no comma
					$decimals = intval($this->nextToken());
					$data_type .= ', ' . $decimals;
					$this->nextToken(); //eat the closing paren explicitly
				}
				$data_type .= ')';
			}
		}
		
		//Read flags and collations. Technically, some flags only apply to certain data types, 
		//but this simplistic parser doesn't care about that. It will happily accept things 
		//like "VACHAR(100) UNSIGNED" and "INT(5) CHARACTER SET utf8". 
		$done = false;
		$character_set = null; $collation = null;
		do {
			$token = strtolower($this->lookAhead());
			switch($token){
				case 'unsigned':
				case 'binary':
				case 'zerofill':
					$this->nextToken(); //Eat the token we're looking at, e.g. "UNSIGNED"
					$$token = true;     //This is safe as we only do it for a couple of known tokens.
					break;
					
				case 'character': 
					$this->nextToken(); //Eat "CHARACTER SET"
					$this->nextToken();
					$character_set = strtolower($this->nextToken());
					break;
					
				case 'collate':
					$this->nextToken(); //Eat "COLLATE"
					$collation = strtolower($this->nextToken());
					break; 
					
				//The rest of the column def. is handled by the caller
				case ',': 
				default:
					$done = true;
			}
			
		} while ( !$done );
		
		//TIMESTAMP columns are NOT NULL by default. 
		//See http://dev.mysql.com/doc/refman/4.1/en/silent-column-changes.html
		if ( ($data_type_name == 'timestamp') && !isset($not_null) ){
			$not_null = true;
		}
		
		return compact('data_type', 'data_type_name', 'unsigned', 'binary', 'zerofill', 
		               'character_set', 'collation', 'enum_values', 'length', 'decimals');
	}
	
	/**
	 * Parse an index definition.
	 * 
	 * @return array
	 */
	function parseIndexDefinition(){
		$token = strtolower($this->nextToken());
		if ( !in_array($token, array('index', 'key')) ){
			//We're dealing with a PRIMARY, UNIQUE, FULLTEXT or other non-basic index. 
			$index_modifier = $token;
			
			switch($index_modifier) {
				case 'primary':
					$name = 'PRIMARY';
					break;
					
				case 'fulltext':
					$index_type = 'FULLTEXT';
					break;
					
			}
			
			//Skip over the INDEX/KEY keyword that may follow the modifier
			$next_token = strtolower($this->lookAhead());
			if ( in_array($next_token, array('index', 'key')) ){
				$this->nextToken();
			}
		}		
		
		//Parse index name, if set
		$token = strtolower($this->nextToken());
		if ( ($token != '(') && ($token != 'using') ){
			if ( $token == '`' ){
				$name = strtolower($this->eatUntilEquals('`', WS_SQL_RETURN_CONCATENATED));
			} else {
				$name = $token;
			}
			$token = strtolower($this->nextToken());
		}
		
		//Parse index type, if specified. The syntax is "USING [BTREE | HASH | RTREE]"
		if ( $token == 'using' ){
			$index_type = strtoupper($this->nextToken());
			$token = strtolower($tihs->nextToken());
		}
		
		//Next, there should be the column list (enclosed in parenthesis)
		if ( $token == '(' ){
			//The format of each entry is "col_name [(length)] [ASC | DESC]"
			//(MySQL ignores the ASC/DESC setting for tables using the MyISAM engine)
			$columns = array();
			do {
				//Column name
				$col_name = $this->parseName();
				
				//Length (optional)
				$length = null;
				$token = $this->lookAhead();
				if ( $token == '(' ){
					$this->nextToken();
					$length = intval($this->eatUntilEquals(')', WS_SQL_RETURN_CONCATENATED));
				}
				
				//Collation, ASC/DESC (optional)
				$collation = null;
				$token = strtolower($this->lookAhead());
				if ( $token == 'asc' || $token == 'desc' ){
					$this->nextToken();
					$collation = $token;
				}
				
				$columns[] = compact('col_name', 'length', 'collation');
				$token = $this->nextToken(); //Either ',' or ')'
			} while ( $token == ',' );
		} else {
			//wtf?
		}
		
		//When no name is specified, MySQL will use the name of the first column in the list as 
		//the index name. If there already exists an index with that name, it will append an 
		//incremental numeric suffix to the column name to generate the index name. This parser 
		//only handles the trivial case, i.e. no suffix. 
		if ( !isset($name) && (isset($columns) && !empty($columns)) ){
			$name = $columns[0]['col_name'];
		}
		
		//Foreign keys would also have a reference definition following the column list,
		//but this parser doesn't support foreign keys. If you give it one, it will choke.
		
		//Eat the comma that separates create definitions
		$separator = $this->lookAhead();
		if ( $separator == ',' ){
			$this->nextToken(); 
		}
		
		return compact('index_modifier', 'name', 'index_type', 'columns');
	}
	
	/**
	 * Parse a MySQL schema object name. The name may be optionally quoted with `backticks`.
	 * 
	 * @return string Either the parsed name (in lower case, no backticks), or NULL. 
	 */
	function parseName(){
		$token = $name = $this->nextToken();
		if ( $token == '`' ){
			$name = $this->eatUntilEquals('`', WS_SQL_RETURN_CONCATENATED); 
		}
		if ( $name ){
			return strtolower($name);
		} else {
			return null;
		}
	}
	
	/**
	 * Parse table options specified in a CREATE TABLE query.
	 * 
	 * @return array
	 */
	function parseTableOptions(){
		$options = array('engine' => null, 'character_set' => null, 'collation' => null, 'comment' => null, 'auto_increment' => null);
		
		$done = false;
		do{
			$token = strtolower($this->nextToken());
			switch($token){
				case ';':      //Found the query separator
					$done = true;
					break;
					
				case 'engine': //ENGINE=Foo
				case 'type':
					$this->nextToken(); //skip the equals sign
					$options['engine'] = $this->nextToken();
					break;
					
				case 'default': //Prefix for "CHARACTER SET", "CHARSET" or "COLLATE".  
					break;
					
				case 'character':
					$this->nextToken(); //skip "SET"
				case 'charset':
					$this->nextToken(); //skip "="
					$options['character_set'] = $this->nextToken();
					break;
					
				case 'collate':
					$this->nextToken(); //skip "="
					$options['collation'] = $this->nextToken();
					break;
					
				case 'auto_increment':
					$this->nextToken(); //skip "="
					$options['auto_increment'] = $this->nextToken();
					break;
					
				case 'comment':
					$this->nextToken(); //skip "="
					$options['comment'] = $this->parseQuotedString();
					break;
					
			}
		} while ( !$done );
		
		return $options;
	}
	
	/**
	 * Split the input string into individual tokens.
	 * 
	 * Uses the $sql field as the input and stores the results in the $tokens field.
	 * Resets the internal token index and populates the (non)whitespace flag array.
	 *  
	 * @return void
	 */
	function tokenize(){
		$regexp = '@
			# general identifiers (keywords, table and column names, etc)
			[\w\d_\.]+ |
			  
			# whitespace				  
			(?P<whitespace>[\s\t\r\n]+) |
			
			# comments
			--\s | 
			\#   |
			/\*  |
			\*/  |  
			  
			# grouping characters (brackets, quotes and so on)
			[`\'"\{\}\[\]\(\)] |
			
			# separators
			[,;] |
			
			# escape character 
			\\\  | 
			 
			# everything else
			.	     
		@sxi';
		
		$this->num_tokens = preg_match_all($regexp, $this->sql, $matches);
		if ( $this->num_tokens > 0 ){
			$this->tokens = $matches[0];
		} else {
			$this->tokens = array();
		}
		
		//Flag (non)whitespace tokens. This lets us greatly speed up eatWhitespace() 
		//at the cost of a bit of extra memory being used.
		$this->_not_whitespace = array();
		foreach($this->tokens as $index => $token){
			$this->_not_whitespace[$index] = ($matches['whitespace'][$index] === '');
		}
		
		//Reset token indices
		$this->index = 0;
		$this->_next_salient_index = -1;
		
		return $this->num_tokens;
	}
	
	/**
	 * Retrieve the next token, ignoring whitespace and SQL comments.
	 * 
	 * @return string Either the next token or NULL if there are no more tokens to return. 
	 */
	function nextToken(){
		if ( $this->_next_salient_index > $this->index ){
			//Each lookAhead()call caches the index of the next useful token. This lets us get there
			//directly without walking over all the interceding whitespace and comments a second time.
			$this->index = $this->_next_salient_index;
		} else {
			while( $this->eatWhitespace() || $this->eatComment() ){};
		}		
		
		if ( $this->index >= $this->num_tokens ){
			return null;
		}
		
		return $this->tokens[$this->index++];
	}
	
	/**
	 * Retrieve the next token without advancing the internal token array index.
	 * 
	 * @uses NaiveSqlParser::nextToken()
	 * 
	 * @return string 
	 */
	function lookAhead(){
		$saved_index = $this->index;
		$token = $this->nextToken();
		$this->_next_salient_index = $this->index-1;
		$this->index = $saved_index;
		
		return $token;
	}
	
	/**
	 * Consume whitespace.
	 * 
	 * @param int $return
	 * @return mixed
	 */
	function eatWhitespace($return = WS_SQL_RETURN_COUNT){
		switch($return){
			case WS_SQL_RETURN_COUNT:
				$consumed = 0;
				for(; $this->index < $this->num_tokens; $this->index++){
					if ( $this->_not_whitespace[$this->index] ){
						break;
					}
					$consumed++;
				}
				break;
			
			case WS_SQL_RETURN_CONCATENATED:
				$consumed = '';
				for(; $this->index < $this->num_tokens; $this->index++){
					if ( $this->_not_whitespace[$this->index] ){
						break;
					}
					$consumed .= $this->tokens[$this->index];
				}
				break;
				
			case WS_SQL_RETURN_ARRAY:
				$consumed = array();
				for(; $this->index < $this->num_tokens; $this->index++){
					if ( $this->_not_whitespace[$this->index] ){
						break;
					}
					$consumed[] = $this->tokens[$this->index];
				}
				break;
				
		}
		
		return $consumed;
	}
	
	/**
	 * Consume an SQL comment.
	 * 
	 * @param int $return
	 * @return mixed
	 */
	function eatComment($return = WS_SQL_RETURN_COUNT){
		if ( $this->index < $this->num_tokens ){
			switch($this->tokens[$this->index]){
				case '-- ';
				case "--\n";
				case "--\r";
				case "--\t";
				case '#':  //One-line comment
					return $this->eatUntilEOL($return);
					break;
					
				case '/*': //C-style multiline comment
					return $this->eatUntilEquals('*/', $return);
					break;
			}
		}
		//Not a comment, or no tokens left
		switch($return){
			case WS_SQL_RETURN_COUNT:
				return 0;
			case WS_SQL_RETURN_CONCATENATED:
				return '';
			case WS_SQL_RETURN_ARRAY:
				return array();
		}	
	}
	
	/**
	 * Parse and return a quoted string.
	 * 
	 * This function uses a very primitive algorithm that does not support escape codes
	 * and other such luxuries. Basically, it will concatenate all tokens up to the next 
	 * quote character and return the resulting string.  
	 * 
	 * @param string $quote Optional. The quote character that encloses the string (' or "). If not specified, the function will use the next token as the quote character. 
	 * @return string
	 */
	function parseQuotedString($quote = null){
		if ( $quote == null ){
			$quote = $this->nextToken();
		}
		//TODO: Proper string parsing
		//See http://dev.mysql.com/doc/refman/4.1/en/string-syntax.html
		
		$string = '';
		for(; $this->index < $this->num_tokens; $this->index++){
			$token = $this->tokens[$this->index];
			switch($token){
				case $quote:
					$this->index++;
					break 2; //Exit both the switch() and the for-loop
					
				default:
					$string .= $token;
			}
		}
		return $string;		
	}
	
	/**
	 * Consume all tokens until the specified terminator token is found.
	 * 
	 * After calling this method, the internal token array index will point
	 * to the token that immediately follows the earliest occurence of
	 * the specified terminator token. In effect, the terminator will be
	 * skipped.
	 * 
	 * The terminator token will not included in the return value. If the 
	 * terminator token is not found the function will consume all remaining 
	 * tokens. 
	 * 
	 * @param string $terminator_token
	 * @param int $return Return format.
	 * @return mixed
	 */
	function eatUntilEquals($terminator_token, $return = WS_SQL_RETURN_COUNT){
		$consumed = array();
		for(; $this->index < $this->num_tokens; $this->index++){
			if ( $this->tokens[$this->index] == $terminator_token ){
				$this->index++;
				break;
			}
			$consumed[] = $this->tokens[$this->index];
		}
		return $this->_prepareConsumedTokens($consumed, $return);
	}
	
	/**
	 * Consume all tokens until a token containing a newline character is encoutered.
	 * 
	 * The token containing the newline will not be included in the return value.
	 * If no newlines are found, the function will consume all remaining tokens.
	 * 
	 * @param int $return Return format.
	 * @return mixed
	 */
	function eatUntilEOL($return = WS_SQL_RETURN_COUNT){
		$consumed = array();
		for(; $this->index < $this->num_tokens; $this->index++){
			if ( strpos($this->tokens[$this->index], "\n") !== false ){
				$this->index++;
				break;
			}
			$consumed[] = $this->tokens[$this->index];
		}
		return $this->_prepareConsumedTokens($consumed, $return);
	}
	
	/**
	 * Attempt to consume exactly the specified sequence of tokens (ignores whitespace and comments). 
	 * 
	 * @param array|string $tokens
	 * @param bool $case_sensitive
	 * @param int $return
	 * @return int|array|string
	 */
	function eatSequence($tokens, $case_sensitive = false, $return = WS_SQL_RETURN_COUNT){
		if ( !is_array($tokens) ){
			$tokens = preg_split('@[\s\r\n]+@', $tokens);
		}
		
		$consumed = array();
		$saved_index = $this->index;
		$success = true;
		foreach($tokens as $i => $token){
			if ( !$case_sensitive ){
				$token = strtolower($token);
			}
			
			$next_token = $this->nextToken();
			if ( !$case_sensitive ){
				$next_token = strtolower($next_token);
			} 
			
			if ( $next_token != $token ){
				$success = false;
				break;
			}			
		}
		
		if ( !$success ){
			$this->index = $saved_index;
			$tokens = array();
		}
		return $this->_prepareConsumedTokens($tokens, $return);
	}
	
	/**
	 * NaiveSqlParser::_prepareConsumedTokens()
	 * 
	 * @access private
	 * 
	 * @param array $tokens
	 * @param mixed $how
	 * @return mixed
	 */
	function _prepareConsumedTokens($tokens, $how = WS_SQL_RETURN_COUNT){
		switch($how) {
			case WS_SQL_RETURN_COUNT:
				return count($tokens);
				break;
				
			case WS_SQL_RETURN_ARRAY:
				return $tokens;
				break;
				
			case WS_SQL_RETURN_CONCATENATED:
				return implode('', $tokens);
				break;
				
			default: 
				return null;
		}
	}
	
	/**
	 * Build an index definition string from its parsed representation
	 * 
	 * This method can be called statically.
	 * 
	 * @param array $parsed_definition Index definition struct. as returned by parseIndexDefinition()
	 * @return string
	 */
	function buildIndexDefinition($parsed_definition){
		$index = 'KEY';
		if ( !empty($parsed_definition['index_modifier']) ){
			$index = strtoupper($parsed_definition['index_modifier']) . ' ' . $index;
		}
		
		//Indexes always have names, even if only implied (exception : primary keys)
		$name = '';
		if ( empty($parsed_definition['name']) ){
			$name = $parsed_definition['columns'][0]['col_name'];
		} elseif ( $parsed_definition['name'] != 'PRIMARY' ){
			$name = $parsed_definition['name'];
		}
		if ( $name ){
			$index .= sprintf(' `%s`', $name);
		}
		
		$column_list = array();
		foreach($parsed_definition['columns'] as $index_column){
			$column = '`' . $index_column['col_name'] . '`';
			if ( !empty($index_column['length']) ){
				$column .= '(' . $index_column['length'] . ')';
			}
			if ( !empty($index_column['collation']) ){
				$column .= ' ' . strtoupper($index_column['collation']);
			}
			$column_list[] = $column;
		}
		
		$index .= ' (' . implode(', ', $column_list) . ')';
						
		return $index;		
	}
	
	/**
	 * Build a column definition string from its parsed representation.
	 * 
	 * This method can be called statically.
	 * 
	 * @param array $parsed_definition Column definition struct. returned by parseColumnDefinition()
	 * @return string
	 */
	function buildColumnDefinition($parsed_definition){
		$column = sprintf('`%s` %s', $parsed_definition['name'], $parsed_definition['data_type']);
		
		if ( array_key_exists('unsigned', $parsed_definition) && $parsed_definition['unsigned'] ){
			$column .= ' UNSIGNED';
		}
		if ( array_key_exists('zerofill', $parsed_definition) && $parsed_definition['zerofill'] ){
			$column .= ' ZEROFILL';
		}
		if ( array_key_exists('binary', $parsed_definition) && $parsed_definition['binary'] ){
			$column .= ' BINARY';
		}	
						
		if ( !empty($parsed_definition['character_set']) ){
			$column .= ' CHARACTER SET '.$parsed_definition['character_set'];
		}
		if ( !empty($parsed_definition['collation']) ){
			$column .= ' COLLATE '.$parsed_definition['collation'];
		}
		if ( isset($parsed_definition['not_null']) ){
			if ( $parsed_definition['not_null'] ){
				$column .= ' NOT NULL';
			} else {
				$column .= ' NULL';
			}
		}
		if ( array_key_exists('default', $parsed_definition) ){
			if ( $parsed_definition['default'] === null ){
				$column .= ' DEFAULT NULL';
			} else {
				$column .= " DEFAULT '".$parsed_definition['default']."'";
			}
		}
		if ( !empty($parsed_definition['auto_increment']) ){
			$column .= ' AUTO_INCREMENT';
		}
		if ( isset($parsed_definition['comment']) ){
			$column .= " COMMENT '" . $parsed_definition['comment'] . "'";
		}
		
		return $column;
	}
}

?>