<?php

class SchemaDiff {
	
	/**
	 * Parse one or more CREATE TABLE queries and generate a list of SQL queries that need 
	 * to be executed to make the current database schema match those queries. Will also
	 * execute those queries if the $execute parameter is set to True.
	 * 
	 * The second argument should be an array listing the types of DDL changes that the function
	 * is allowed to perform. It may contain any of the following values: 
	 *  'create_table'    - create new tables.
	 *  'add_column'      - add new columns to existing tables.
	 *  'drop_column'     - delete columns that are not present in the new schema.
	 *  'modify_column'   - modify the data type and attributes of existing columns.
	 *  'add_index'       - add new indexes to existing tables. 
	 *  'drop_index'      - delete indexes not present in the new schema.
	 *  'execute_unknown' - blindly execute any input query that is not a CREATE TABLE query.
	 * 
	 * Note the absence of a 'modify_index' operation. Indexes can't be modified; only dropped 
	 * and recreated. This will be done automatically if the 'add_index' and 'drop_index'
	 * operations are both allowed.
	 * 
	 * By default, all operations are allowed.
	 * 
	 * If $execute is False (the default), this function will return an array of strings 
	 * where each string is a DDL query. If it True, it will instead execute each generated 
	 * query and return an array where each item is an assoc. array with these keys :
	 *  'query'         - the executed SQL query.
	 *  'success'       - True if the query was executed successfully, False if it caused an error.
	 *  'error_message' - the DB error message (only meaningful when 'success' = false).
	 * 
	 * @param string $queries One or more CREATE TABLE queries separated by a semicolon.
	 * @param array $allowed_operations List of DDL modifications that the function is allowed to make.
	 * @param bool $execute Whether to apply the schema changes. Defaults to false.
	 * @return array Either an array of DDL queries, or an array of arrays containing the queries and their execution results.  
	 */
	function diff($queries, $allowed_operations = null, $execute = false){
		if ( $allowed_operations == null ){
			$allowed_operations = array('create_table', 'add_column', 'drop_column', 'modify_column', 'add_index', 'drop_index', 'execute_unknown');
		}
		
		$parser = new NaiveSqlParser($queries);
		$parsed_queries = $parser->parse();
		$queries_to_run = array();
				
		foreach($parsed_queries as $query){
			
			if ( $query['query_type'] == 'create_table' ){
				$current_create_query = $this->getCreateQuery($query['name']);
				if ( $current_create_query ) {
					$current_schema = $parser->parse($current_create_query);
					$current_schema = reset($current_schema);

					$schema_update_queries = $this->tableDiff($current_schema, $query, $allowed_operations);
					$queries_to_run = array_merge($queries_to_run, $schema_update_queries);
				} else {
					//The table doesn't exist, so it will be created
					if ( in_array('create_table', $allowed_operations) ){
						$queries_to_run[] = $query['raw_query'];
					}					
				}				
			} else {
				if ( in_array('execute_unknown', $allowed_operations) ){
					$queries_to_run[] = $query['raw_query'];
				}
			}
		}
		
		if ( $execute ){
			$results = array();
			foreach($queries_to_run as $query){
				$execution_result = $this->executeQuery($query);
				$execution_result['query'] = $query;
				$results[] = $execution_result;
			}
			return $results;
		}
		
		return $queries_to_run;
	}
	
	/**
	 * SchemaDiff::tableDiff()
	 * 
	 * @param array $base_schema
	 * @param array $new_schema
	 * @return array of DDL queries
	 */
	function tableDiff($base_schema, $new_schema, $allowed_operations){
		$queries = array();
				
		$new_columns = $new_schema['columns'];
		$old_columns = $base_schema['columns'];
		
		foreach($old_columns as $name => $old_column){
			//Is this column still present in the new schema?
			if ( array_key_exists($name, $new_columns) ){
				$new_column = $new_columns[$name];
				unset($new_columns[$name]); //This column doesn't need to be added
				
				if ( !in_array('modify_column', $allowed_operations) ){
					continue;
				}
								
				//Has the data type or other attributes changed in the new schema?
				$modify =
					( $new_column['data_type'] != $old_column['data_type'] ) ||
					( empty($new_column['unsigned']) != empty($old_column['unsigned']) ) ||
					( empty($new_column['zerofill']) != empty($old_column['zerofill']) ) ||
					( empty($new_column['binary']) != empty($old_column['binary']) )     ||
					( empty($new_column['not_null']) != empty($old_column['not_null']) ) ||
					( empty($new_column['auto_increment']) != empty($old_column['auto_increment']) );
					
				/*
				Character sets and collation are a bit more complex because we have to take
				the table-wide defaults into account. 
				
				Note that the charset <=> collation correspondence is ignored. This means that
				even though charset is implied by the collation (and vice versa) you need to 
				specify both for this function to reliably detect changes in these attributes.			
				*/
				$new_character_set = !empty($new_column['character_set']) ? $new_column['character_set'] : $new_schema['options']['character_set'];
				$new_collation = !empty($new_column['collation']) ? $new_column['collation'] : $new_schema['options']['collation'];
				$old_character_set = !empty($old_column['character_set']) ? $old_column['character_set'] : $base_schema['options']['character_set'];
				$old_collation = !empty($old_column['collation']) ? $old_column['collation'] : $base_schema['options']['collation'];
				
				$modify = $modify ||
					( !empty($new_character_set) && ($new_character_set != $old_character_set) ) ||
					( !empty($new_collation) && ($new_collation != $old_collation) );
				
				if ( $modify ){
					$queries[] = sprintf(
						'ALTER TABLE `%s` MODIFY COLUMN %s',
						$new_schema['name'],
						NaiveSqlParser::buildColumnDefinition($new_column)
					);
				}
			} else {
				//Drop obsolete columns
				if ( in_array('drop_column', $allowed_operations) ){
					$queries[] = sprintf(
						'ALTER TABLE `%s` DROP COLUMN `%s`',
						$new_schema['name'],
						$name
					);
				}
			}
		}
		
		//Add columns that don't exist yet 
		if ( !empty($new_columns ) && in_array('add_column', $allowed_operations) ){
			foreach($new_columns as $column){
				$queries[] = sprintf(
					'ALTER TABLE `%s` ADD COLUMN %s',
					$new_schema['name'],
					NaiveSqlParser::buildColumnDefinition($column)
				);
			}
		}
		
		return $queries;
	}
	
	/**
	 * SchemaDiff::getCreateQuery()
	 * 
	 * @param string $table_name
	 * @return string The CREATE TABLE query that creates the specified table, or FALSE if the table doesn't exist.
	 */
	function getCreateQuery($table_name){
		return false;
	}
	
	/**
	 * SchemaDiff::executeQuery()
	 * 
	 * @param string $query
	 * @return array Associative array with the keys 'success' and 'error_message'.
	 */
	function executeQuery($query){
		return array('success' => false, 'error_message' => 'Not implemented');
	}		
}

?>